"""

The database migration is based on alembic_. The usage is:

.. code-block:: text

    usage: owl-migrate [-h] [-c CONF]
                           {current,edit,show,history,revision,upgrade} ...

    Owl database migration

    optional arguments:
      -h, --help            show this help message and exit

    command:
      {current,edit,show,history,revision,upgrade}
        current             Display current revision
        edit                Edit revision
        show                Show revision
        history             Show revision history
        revision            Create a new revision file
        upgrade             Upgrade to a later version


.. _alembic: http://alembic.zzzcomputing.com/en/latest/

"""

import argparse
from datetime import datetime

from alembic import command as alembic_commands
from alembic.config import Config
from owl_server.config import config


def main():
    main_parser = argparse.ArgumentParser(description="Owl database migration")
    cmd_subparsers = main_parser.add_subparsers(title="command", dest="command")

    # current
    current_parser = cmd_subparsers.add_parser(
        "current", help="Display current revision"
    )
    current_parser.add_argument(
        "-v", "--verbose", action="store_true", help="More verbose."
    )

    # edit
    edit_parser = cmd_subparsers.add_parser("edit", help="Edit revision")
    edit_parser.add_argument("revision", type=str, help="Revision")

    # show
    show_parser = cmd_subparsers.add_parser("show", help="Show revision")
    show_parser.add_argument("revision", type=str, help="Revision")

    # history
    history_parser = cmd_subparsers.add_parser("history", help="Show revision history")
    history_parser.add_argument(
        "-v", "--verbose", action="store_true", help="More verbose."
    )

    # revision
    revision_parser = cmd_subparsers.add_parser(
        "revision", help="Create a new revision file"
    )
    revision_parser.add_argument("-m", "--message", type=str, help="Revision message.")
    revision_parser.add_argument("--id", type=str, default=None, help="Revision id.")
    revision_parser.add_argument(
        "--autogenerate", action="store_true", help="Auto generate SQL from database"
    )

    # upgrade
    upgrade_parser = cmd_subparsers.add_parser(
        "upgrade", help="Upgrade to a later version"
    )
    upgrade_parser.add_argument("revision", type=str, help="Revision")
    upgrade_parser.add_argument("--sql", action="store_true", help="SQL mode.")

    args = main_parser.parse_args()

    alembic_cfg = Config()
    alembic_cfg.set_main_option("script_location", "owl_server:migrations")
    alembic_cfg.set_main_option("sqlalchemy.url", config.dbi)

    if args.command == "current":
        alembic_commands.current(alembic_cfg, verbose=args.verbose)
    elif "edit" in args.command:
        alembic_commands.edit(alembic_cfg, args.revision)
    elif "show" in args.command:
        alembic_commands.show(alembic_cfg, args.revision)
    elif "history" in args.command:
        alembic_commands.history(alembic_cfg, verbose=args.verbose)
    elif "revision" in args.command:
        if args.id is None:
            rev_id = datetime.now().strftime("%Y%m%d")
        else:
            rev_id = args.id
        alembic_commands.revision(
            alembic_cfg,
            message=args.message,
            rev_id=rev_id,
            autogenerate=args.autogenerate,
        )
    elif "upgrade" in args.command:
        alembic_commands.upgrade(alembic_cfg, args.revision, sql=args.sql)



import logging
import random
import string

from kubernetes_asyncio import client, config
from kubernetes_asyncio.client.api_client import ApiClient
from kubernetes_asyncio.client.rest import ApiException
from kubernetes_asyncio.config.config_exception import \
    ConfigException  # noqa: F401

from .utils import _make_spec_from_dict

log = logging.getLogger("owl.daemon.scheduler")


# Setup K8 configs
# config.load_kube_config()
# config.load_incluster_config()
# configuration = client.Configuration()
# api_instance = client.BatchV1Api()


def kube_delete_empty_pods(namespace=None, phase=None):
    """
    Pods are never empty, just completed the lifecycle.
    As such they can be deleted.
    Pods can be without any running container in 2 states:
    Succeeded and Failed. This call doesn't terminate Failed pods by default.
    """
    namespace = namespace or "default"
    phase = phase or "Succeeded"
    # The always needed object
    deleteoptions = client.V1DeleteOptions()
    # We need the api entry point for pods
    api_pods = client.CoreV1Api()
    # List the pods
    try:
        pods = api_pods.list_namespaced_pod(namespace, pretty=True, timeout_seconds=60)
    except ApiException as e:
        logging.error("Exception when calling CoreV1Api->list_namespaced_pod: %s\n" % e)

    for pod in pods.items:
        logging.debug(pod)
        podname = pod.metadata.name
        try:
            if pod.status.phase == phase:
                api_response = api_pods.delete_namespaced_pod(
                    podname, namespace, deleteoptions
                )
                logging.info("Pod: {} deleted!".format(podname))
                logging.debug(api_response)
            else:
                logging.info(
                    "Pod: {} still not done... Phase: {}".format(
                        podname, pod.status.phase
                    )
                )
        except ApiException as e:
            logging.error(
                "Exception when calling CoreV1Api->delete_namespaced_pod: %s\n" % e
            )

    return


def kube_create_job_object(
    name,
    container_image,
    command=None,
    namespace=None,
    container_name=None,
    env_vars=None,
    extraConfig=None,
    service_account_name=None,
):
    """
    Create a k8 Job Object
    Minimum definition of a job object:
    {'api_version': None, - Str
    'kind': None,     - Str
    'metadata': None, - Metada Object
    'spec': None,     -V1JobSpec
    'status': None}   - V1Job Status
    Docs: https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Job.md
    Docs2: https://kubernetes.io/docs/concepts/workloads/controllers/jobs-run-to-completion/#writing-a-job-spec
    Also docs are pretty pretty bad. Best way is to ´pip install kubernetes´ and go via the autogenerated code
    And figure out the chain of objects that you need to hold a final valid object So for a job object you need:
    V1Job -> V1ObjectMeta
          -> V1JobStatus
          -> V1JobSpec -> V1PodTemplate -> V1PodTemplateSpec -> V1Container

    Now the tricky part, is that V1Job.spec needs a .template, but not a PodTemplateSpec, as such
    you need to build a PodTemplate, add a template field (template.template) and make sure
    template.template.spec is now the PodSpec.
    Then, the V1Job.spec needs to be a JobSpec which has a template the template.template field of the PodTemplate.
    Failure to do so will trigger an API error.
    Also Containers must be a list!
    Docs3: https://github.com/kubernetes-client/python/issues/589
    """
    env_vars = env_vars or {}
    namespace = namespace or "default"
    command = command or "sleep 60"
    container_name = container_name or "jobcontainer"
    # Body is the object Body
    body = client.V1Job(api_version="batch/v1", kind="Job")
    # Body needs Metadata
    # Attention: Each JOB must have a different name!
    body.metadata = client.V1ObjectMeta(namespace=namespace, name=name)
    # And a Status
    body.status = client.V1JobStatus()
    # Now we start with the Template...
    template = client.V1PodTemplate()
    template.template = client.V1PodTemplateSpec()
    # Passing Arguments in Env:
    env_list = []
    for env_name, env_value in env_vars.items():
        env_list.append(client.V1EnvVar(name=env_name, value=str(env_value)))

    if (resources := extraConfig.resources) is not None:
        resourcesReq = client.V1ResourceRequirements(**resources)
    else:
        resourcesReq = None

    if (security_context := extraConfig.securityContext) is not None:
        secReq = _make_spec_from_dict(security_context, client.V1PodSecurityContext)
    else:
        secReq = None

    volume_mounts = []
    for vm in extraConfig.volumeMounts:
        volume_mounts.append(_make_spec_from_dict(vm, client.V1VolumeMount))

    container = client.V1Container(
        name=container_name,
        image=container_image,
        env=env_list,
        image_pull_policy="IfNotPresent",
        args=command.split(),
        resources=resourcesReq,
        volume_mounts=volume_mounts,
    )

    volumes = []
    for vol in extraConfig.volumes:
        volumes.append(_make_spec_from_dict(vol, client.V1Volume))

    template.template.spec = client.V1PodSpec(
        containers=[container],
        restart_policy="Never",
        termination_grace_period_seconds=30,
        volumes=volumes,
        service_account_name=service_account_name,
        security_context=secReq,
    )

    # And finaly we can create our V1JobSpec!
    body.spec = client.V1JobSpec(
        ttl_seconds_after_finished=300, template=template.template, backoff_limit=2
    )
    return body


async def kube_create_job(
    name,
    image,
    command=None,
    namespace=None,
    extraConfig=None,
    env_vars=None,
    service_account_name=None,
):
    namespace = namespace or "default"
    # Create the job definition
    body = kube_create_job_object(
        name,
        image,
        env_vars=env_vars,
        namespace=namespace,
        command=command or "sleep 60",
        extraConfig=extraConfig or {},
        service_account_name=service_account_name,
    )

    config.load_incluster_config()
    async with ApiClient() as api:
        v1 = client.BatchV1Api(api)
        res = await v1.create_namespaced_job(namespace, body, pretty=True)
    return res


async def kube_delete_job(name, namespace=None):
    config.load_incluster_config()
    async with ApiClient() as api:
        v1 = client.BatchV1Api(api)
        res = await v1.delete_namespaced_job(
            name, namespace or "default", propagation_policy="Background", grace_period_seconds=0
        )
    return res


def id_generator(size=12, chars=string.ascii_lowercase + string.digits):
    return "".join(random.choice(chars) for _ in range(size))


async def kube_test_credentials():
    """
    Testing function.
    If you get an error on this call don't proceed. Something is wrong on your connectivty to
    Google API.
    Check Credentials, permissions, keys, etc.
    Docs: https://cloud.google.com/docs/authentication/
    """
    config.load_incluster_config()

    async with ApiClient() as api:
        v1 = client.BatchV1Api(api)
        await v1.get_api_resources()



from distributed import WorkerPlugin
from loky import ProcessPoolExecutor

# from concurrent.futures import ProcessPoolExecutor


class LoggingPlugin(WorkerPlugin):
    def __init__(self, config):
        self.config = config

    def setup(self, worker):
        import logging.config

        logging.config.dictConfig(self.config)


class ProcessPoolPlugin(WorkerPlugin):
    def setup(self, worker):
        executor = ProcessPoolExecutor(max_workers=worker.nthreads)
        worker.executors["processes"] = executor



import asyncio
import functools
import json
import os
from contextlib import suppress
from datetime import datetime
from typing import Any, Dict, Optional

import databases
import zmq
import zmq.asyncio
from fastapi import FastAPI, Header, HTTPException, status
from owl_server.config import config
from pydantic import BaseModel

from .. import database as db
from ..crypto import PBKDF2PasswordHasher, get_random_string

OWL_USERNAME = "owl"

database = databases.Database(config.dbi)


def authenticate(admin=False):
    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            auth = kwargs["authentication"]
            auth += " dummy dummy"
            username, token, *_ = auth.split()
            await check_token(username, token)
            if admin:
                await check_admin(username)
            kwargs["username"] = username
            return await func(*args, **kwargs)

        return wrapped

    return wrapper


async def check_token(username, token) -> bool:
    q = db.Token.select().where(db.Token.c.username == username)
    res = await database.fetch_one(q)
    if (not res) or (res["token"] != token):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized",
        )


async def check_password(user) -> bool:
    q = db.User.select().where(db.User.c.username == user.username)
    res = await database.fetch_one(q)
    if not res:
        return False
    hasher = PBKDF2PasswordHasher()
    return hasher.verify(user.password, res["password"])


async def check_admin(username) -> bool:
    q = db.User.select().where(db.User.c.username == username)
    res = await database.fetch_one(q)
    if (not res) or (res["is_admin"] is False):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized",
        )


async def check_config(config):
    await asyncio.sleep(0)
    if not config:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail="Empty config",
        )
    return config


async def check_status(st):
    await asyncio.sleep(0)
    if st not in [
        "ALL",
        "PENDING",
        "STARTING",
        "RUNNING",
        "FINISHED",
        "ERROR",
        "TO_CANCEL",
        "CANCELLED",
    ]:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail="Invalid status",
        )
    return st


def AdminSocket():
    host = os.environ.get("OWL_SCHEDULER_SERVICE_HOST")
    port = os.environ.get("OWL_SCHEDULER_SERVICE_PORT_ADMIN")
    admin_addr = f"tcp://{host}:{port}"
    ctx = zmq.asyncio.Context()
    admin_socket = ctx.socket(zmq.DEALER)
    admin_socket.setsockopt(zmq.IDENTITY, config.token.encode("utf-8"))
    admin_socket.connect(admin_addr)
    return admin_socket


app = FastAPI()

admin_socket = AdminSocket()


@app.on_event("startup")
async def startup():
    await database.connect()
    token = config.token
    try:
        q = db.Token.insert().values(username=OWL_USERNAME, token=token)
        await database.execute(q)
    except Exception:
        q = (
            db.Token.update()
            .where(db.Token.c.username == OWL_USERNAME)
            .values(token=token)
        )
        await database.execute(q)


@app.on_event("shutdown")
async def shutdown():
    await database.disconnect()


class User(BaseModel):
    username: str
    password: Optional[str] = None
    is_admin: Optional[bool] = False
    active: Optional[bool] = True


class Pipeline(BaseModel):
    config: Optional[Dict[str, Any]] = None
    status: Optional[str] = None


class PipeDef(BaseModel):
    name: str
    pdef: str
    extra_pip_packages: str
    active: Optional[bool] = True


class AdminCommand(BaseModel):
    cmd: Dict[str, Any]


@app.get("/api/pipeline/list/{status}")
@authenticate()
async def pipeline_list(
    status: str, listall: bool = False, authentication=Header(None), username=None
):
    st = await check_status(status.upper())
    q = db.Pipeline.select()
    if st not in ["ALL"]:
        q = q.where(db.Pipeline.c.status == st)
    if username not in [OWL_USERNAME]:
        if listall:
            try:
                await check_admin(username)
            except HTTPException:
                q = q.where(db.Pipeline.c.user == username)
        else:
            q = q.where(db.Pipeline.c.user == username)
    q = q.order_by(db.Pipeline.c.id.desc())
    res = await database.fetch_all(q)
    return res


@app.get("/api/pipeline/status/{uid}")
@authenticate()
async def pipeline_status_get(uid: int, authentication=Header(None), username=None):
    q = db.Pipeline.select().where(db.Pipeline.c.id == uid)
    res = await database.fetch_one(q)
    return res


@app.get("/api/pipeline/log/{uid}")
@authenticate()
async def pipeline_log(uid: int, authentication=Header(None), username=None):
    # TODO: check that pipeline user == username
    await asyncio.sleep(0)
    log = ""
    with suppress(Exception):
        log = open(f"/var/run/owl/logs/pipeline_{uid}.log").read()
    return {"log": log}


@app.post("/api/pipeline/add")
@authenticate()
async def pipeline_add(pipe: Pipeline, authentication=Header(None), username=None):
    config = await check_config(pipe.config)
    q = db.Pipeline.insert().values(
        user=username, config=config, created_at=datetime.now(), status="PENDING"
    )
    uid = await database.execute(q)
    return {"id": uid}


@app.post("/api/pipeline/cancel/{uid}")
@authenticate()
async def pipeline_cancel(uid: int, authentication=Header(None), username=None):
    q = db.Pipeline.select().where(db.Pipeline.c.id == uid)
    res = await database.fetch_one(q)
    if res["user"] != username:
        q = db.User.select().where(db.User.c.username == username)
        res = await database.fetch_one(q)
        if not res["is_admin"]:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED, detail="Unauthorized",
            )
    q = db.Pipeline.update().where(db.Pipeline.c.id == uid).values(status="TO_CANCEL")
    await database.execute(q)
    return {"id": uid, "status": "TO_CANCEL"}


@app.post("/api/pipeline/update/{uid}")
@authenticate()
async def pipeline_update(
    uid: int, pipe: Pipeline, authentication=Header(None), username=None
):
    new = await check_status(pipe.status.upper())
    q = db.Pipeline.select().where(db.Pipeline.c.id == uid)
    res = await database.fetch_one(q)
    if not res:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f"Pipeline {uid} not found",
        )
    if username not in [OWL_USERNAME, res["user"]]:
        await check_admin(username)
    update = False
    if (res["status"] in ["TO_CANCEL"]) and (new == "CANCELLED"):
        update = True
    elif res["status"] not in ["TO_CANCEL"]:
        update = True
    if update:
        q = db.Pipeline.update().where(db.Pipeline.c.id == uid).values(status=new)
        await database.execute(q)

    return {"id": uid, "status": new, "user": res["user"]}


@app.post("/api/admin/command")
@authenticate(admin=True)
async def admin_command(
    command: AdminCommand, authentication=Header(None), username=None
):
    msg = json.dumps(command.cmd).encode("utf-8")
    await admin_socket.send(msg)
    return {"status": "message sent"}


@app.post("/api/auth/login")
async def login(user: User):
    res = await check_password(user)
    if not res:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
        )
    token = get_random_string(64)
    try:
        q = db.Token.insert().values(username=user.username, token=token)
        await database.execute(q)
    except Exception:
        q = (
            db.Token.update()
            .where(db.Token.c.username == user.username)
            .values(token=token)
        )
        await database.execute(q)
    return {"token": token}


@app.post("/api/auth/logout")
@authenticate()
async def logout(user: User, authentication=Header(None), username=None):
    q = db.Token.delete().where(db.Token.c.username == username)
    with suppress(Exception):
        await database.execute(q)
    return {"user": user.username}


@app.post("/api/auth/user/add")
@authenticate(admin=True)
async def add_user(user: User, authentication=Header(None), username=None):
    hasher = PBKDF2PasswordHasher()
    salt = hasher.salt()
    password = hasher.encode(user.password, salt)
    try:
        q = db.User.insert().values(
            username=user.username,
            password=password,
            is_admin=user.is_admin,
            active=user.active,
        )
        await database.execute(q)
    except Exception:
        raise HTTPException(
            status_code=status.HTTP_208_ALREADY_REPORTED, detail="User already exists",
        )
    return {"user": user.username}


@app.post("/api/auth/user/update")
@authenticate(admin=True)
async def update_user(user: User, authentication=Header(None), username=None):
    data = {"is_admin": user.is_admin, "active": user.active}
    if user.password:
        hasher = PBKDF2PasswordHasher()
        salt = hasher.salt()
        password = hasher.encode(user.password, salt)
        data.update({"password": password})

    q = db.User.update().where(db.User.c.username == user.username).values(**data)
    await database.execute(q)

    if "password" in data:
        q = db.Token.delete(db.Token.c.username == user.username)
        await database.execute(q)

    return {"user": user.username}


@app.post("/api/auth/change_password")
@authenticate()
async def change_password(user: User, authentication=Header(None), username=None):
    if not user.password:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Password not supplied",
        )

    hasher = PBKDF2PasswordHasher()
    salt = hasher.salt()
    password = hasher.encode(user.password, salt)
    data = {"password": password}

    q = db.User.update().where(db.User.c.username == username).values(**data)
    await database.execute(q)

    q = db.Token.delete(db.Token.c.username == username)
    await database.execute(q)

    return {"user": user.username}


@app.get("/api/auth/user/get/{user}")
@authenticate(admin=True)
async def get_user(user: str, authentication=Header(None), username=None):
    q = db.User.select()
    if user != "0":
        q = q.where(db.User.c.username == user)
    else:
        q = q.order_by(db.User.c.id)
    res = await database.fetch_all(q)
    # remove password
    resd = [dict(r) for r in res]
    [r.pop("password", None) for r in resd]
    return resd


@app.post("/api/auth/user/delete")
@authenticate(admin=True)
async def delete_user(user: User, authentication=Header(None), username=None):
    q = db.User.delete().where(db.User.c.username == user.username)
    await database.execute(q)

    q = db.Token.delete(db.Token.c.username == user.username)
    await database.execute(q)

    return {"user": user.username}


@app.post("/api/pdef/add")
@authenticate(admin=True)
async def add_pdef(pdef: PipeDef, authentication=Header(None), username=None):
    try:
        q = db.PipelineDefinition.insert().values(
            name=pdef.name,
            pdef=pdef.pdef,
            extra_pip_packages=pdef.extra_pip_packages,
            active=pdef.active,
        )
        await database.execute(q)
        action = "created"
    except Exception:
        q = (
            db.PipelineDefinition.update()
            .where(db.PipelineDefinition.c.name == pdef.name)
            .values(
                pdef=pdef.pdef,
                extra_pip_packages=pdef.extra_pip_packages,
                active=pdef.active,
            )
        )
        await database.execute(q)
        action = "updated"
    return {"name": pdef.name, "action": action}


@app.get("/api/pdef/list")
@authenticate()
async def list_pdef(authentication=Header(None), username=None):
    q = db.PipelineDefinition.select().order_by(db.PipelineDefinition.c.id)
    res = await database.fetch_all(q)
    return res


@app.get("/api/pdef/get/{name}")
@authenticate()
async def get_pdef(name: str, authentication=Header(None), username=None):
    q = db.PipelineDefinition.select().where(db.PipelineDefinition.c.name == name)
    res = await database.fetch_one(q)
    return res


@app.get("/")
def read_root():
    return {"Hello": "World"}



import asyncio
import logging
import logging.config
import signal
from argparse import Namespace

from owl_server.config import config  # noqa: F401
from owl_server.daemon import Pipeline
from owl_server.log import initlog
from owl_server.schema import schema_pipeline
from owl_server.utils import read_config

logger = logging.getLogger("owl.cli")


def run_pipeline(args: Namespace) -> None:
    """Run scheduler.

    Parameters
    ----------
    arg
        Argparse namespace containing command line flags.
    """
    initlog("pipeline")

    pdef = read_config(args.conf, schema_pipeline)

    try:
        start_loop(pdef)
    except:  # noqa
        logger.critical("Error running the pipeline", exc_info=True)
        raise

    logger.info("Pipeline run ended.")


def start_loop(pdef):
    loop = asyncio.get_event_loop()
    logger.info("Starting Pipeline")
    pipeline = Pipeline(pdef)

    loop.add_signal_handler(signal.SIGTERM, pipeline.close)
    loop.add_signal_handler(signal.SIGINT, pipeline.close)

    res = loop.run_until_complete(pipeline.run())
    loop.close()

    if res in ["ERROR"]:
        raise Exception()
